package org.confluence.mod.common.item.food;

import net.minecraft.ChatFormatting;
import net.minecraft.network.chat.Component;
import net.minecraft.server.level.ServerLevel;
import net.minecraft.sounds.SoundEvent;
import net.minecraft.sounds.SoundEvents;
import net.minecraft.world.entity.LivingEntity;
import net.minecraft.world.food.FoodProperties;
import net.minecraft.world.item.Item;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.item.TooltipFlag;
import net.minecraft.world.item.UseAnim;
import net.minecraft.world.level.Level;
import org.confluence.mod.common.init.ModSecretSeeds;
import org.confluence.mod.common.init.item.FoodItems;

import java.util.ArrayList;
import java.util.List;
import java.util.function.Function;

public class BaseFoodItem extends Item {
    protected Builder builder;

    public BaseFoodItem(Builder builder) {
        super(builder.properties);
        this.builder = builder;
    }

    public static Builder builder() {
        return new Builder(new Properties());
    }

    @Override
    public int getUseDuration(ItemStack stack, LivingEntity entity) {
        return builder.duration.apply(stack);
    }

    @Override
    public UseAnim getUseAnimation(ItemStack stack) {
        return builder.useAnim.apply(stack);
    }

    @Override
    public SoundEvent getDrinkingSound() {
        return builder.drinkingSoundType.apply(null);
    }

    @Override
    public SoundEvent getEatingSound() {
        return builder.eatingSoundType.apply(null);
    }

    @Override
    public ItemStack finishUsingItem(ItemStack stack, Level level, LivingEntity livingEntity) {
        if (stack.is(FoodItems.CHERRY) && level instanceof ServerLevel serverLevel && ModSecretSeeds.NO_TRAPS.match(serverLevel) && level.random.nextFloat() < 0.7F) {
            level.explode(null, livingEntity.getX(), livingEntity.getY(), livingEntity.getZ(), 2.5F, false, Level.ExplosionInteraction.MOB);
        }
        return super.finishUsingItem(stack, level, livingEntity);
    }

    @Override
    public void appendHoverText(ItemStack stack, TooltipContext context, List<Component> tooltipComponents, TooltipFlag tooltipFlag) {
        tooltipComponents.addAll(builder.tooltips);
        super.appendHoverText(stack, context, tooltipComponents, tooltipFlag);
    }

    public static class Builder {
        private final Properties properties;
        private Function<ItemStack, Integer> duration = duration -> 0;
        private Function<Void, SoundEvent> drinkingSoundType = sound -> SoundEvents.EMPTY;
        private Function<Void, SoundEvent> eatingSoundType = sound -> SoundEvents.EMPTY;
        private Function<ItemStack, UseAnim> useAnim = useAnim -> UseAnim.NONE;
        private final List<Component> tooltips = new ArrayList<>();

        Builder(Properties properties) {
            this.properties = properties;
        }

        public void isFireResistant() {
            this.properties.fireResistant();
        }

        public Builder food(FoodProperties foodProperties) {
            properties.food(foodProperties);
            return this;
        }

        public Builder craftRemainder(Item item) {
            properties.craftRemainder(item);
            return this;
        }

        public Builder duration(Function<ItemStack, Integer> duration) {
            this.duration = duration;
            return this;
        }

        public Builder useAnim(Function<ItemStack, UseAnim> useAnim) {
            this.useAnim = useAnim;
            return this;
        }

        public Builder drinkingSound(Function<Void, SoundEvent> drinkingSoundType) {
            this.drinkingSoundType = drinkingSoundType;
            return this;
        }

        public Builder eatingSound(Function<Void, SoundEvent> eatingSoundType) {
            this.eatingSoundType = eatingSoundType;
            return this;
        }

        public Builder tooltip(String id, int lineCount) {
            this.tooltips.addAll(tooltips(id, lineCount));
            return this;
        }

        public static List<Component> tooltips(String id, int lineCount) {
            List<Component> components = new ArrayList<>();
            for (int i = 1; i <= lineCount; i++) {
                components.add(Component.translatable("item.confluence." + id + ".tooltip." + i).withStyle(ChatFormatting.GRAY));
            }
            return components;
        }

        public Builder initialize() {
            properties.stacksTo(64);
            return this;
        }

        public BaseFoodItem build() {
            return new BaseFoodItem(initialize());
        }
    }
}
